
	/* test some of the numerical recipes routines for solving
	   systems of simultaneous equations. */

#include <stdio.h>
#include <math.h>
#include "solvepht.h"
#include "nr.h"

#define TOLERANCE    1.0e-7
#define TINY         (2.0*TOLERANCE) 
#define MAG_MATCH    0.30		/* used by find_sigm() */


	/* we read in data for ee images and ss stars,
	   and form an equation of form 
	                 (A) x = b
	   where 'A' is a matrix, 'b' is a vector with the observed values of
	   instr. mags and weights, and 'x' is the unknown vector of
	   magnitude adjustments that minimizes the sums of the squares
	   of the residuals from the mean. */

static void init_matrix( /* int ee, ss */ );
static void calc_em( /* int m, n */ );
static void calc_mags( /* int m, n */ );
static void calc_sigmas( /* int m, n */ );
static void zero_vector( /* float *p; int m; float tolerance */ );
static void print_vector( /* float *p, int m */ );
static void print_matrix( /* float **p; int m, n */ );
static void fill_bvec( /* float *b; int m, n */ );
static void read_data( /* int m, n */ );
static void fill_matrix( /* float **p; int m, n */ );
static void copy_matrix( /* float **p, **q; int m, n */ );
static void normalize( /* int m, n; double value */ );
static void find_means( /* int m, n */ );
static float find_sigm( /* int n, e, s */ );

int write_images( /* char *file */ );
int write_stars( /* char *file */ );
int write_sigma( /* char *file */ );


	/* these two are inputs */
static float **mag;				/* ee by ss matrix of instr. mags */
static float **weight;			/* ee by ss matrix of weights */
	/* these are computational values */
static float **a;				/* ee+ss by ee+ss matrix; A above */
static float **u;				/* ee+ss by ee+ss: part 1 of the SVD output */
static float *w;				/* ee+ss vector: part 2 of SVD output */
static float **v;				/* ee+ss by ee+ss: part 3 of SVD output */
static float *b;				/* ee+ss vector: 'b' above */
static float *x;				/* ee+ss vector: 'x' above, the 'solution' */
	/* and these are output results */
static float *em;				/* ee vector: mean correction for image e */
static float *mag0;				/* ss vector: mean mag of star s */
static float *sigm0;			/* ss vector: sigma of mean value of star s */
static float **Mes;				/* ee by ss matrix: solved mag of star s */
                   				/*    on image e */
static float **sigMes;			/* ee by ss matrix: total sigma of */
                  				/*    solution's value of star s on image e */
static float *sigem;			/* ee vector: sigma of exp corr. of image e */
static float *sigemn;			/* ee vector: sigem/# stars in image e */

solve_matrix(zero_point)
double zero_point;
{
	int ee, ss;
	float tolerance;

	tolerance = TOLERANCE;
	ee = num_images;
	ss = num_ids;
	init_matrix(ee, ss);

	read_data(ee, ss);
	fill_matrix(a, ee, ss);
	if (debug_flag) {
		printf("\n This is the matrix \n");
		print_matrix(a, ee + ss, ee + ss);
	}
	copy_matrix(a, u, ee + ss, ee + ss);

	fill_bvec(b, ee, ss);
	if (debug_flag) {
		printf(" This is the B vector \n");
		print_vector(b, ee + ss);
	}

	svdcmp(u, ee + ss, ee + ss, w, v);

	if (debug_flag) {
		printf("this is the svd matrix U \n");
		print_matrix(u, ee + ss, ee + ss);
		printf("this is the vector of w(j) - which could need zero'ing \n");
		print_vector(w, ee + ss);
	}

	zero_vector(w, ee + ss, tolerance);
	if (debug_flag) {
		printf("this is the vector w(j) after zeroing tiny elements\n");
		print_vector(w, ee + ss);
	}

	svbksb(u, w, v, ee + ss, ee + ss, b, x);

	if (debug_flag) {
		printf("\nand this is the vector of x - the solution\n");
		print_vector(x, ee + ss);
	}

	calc_em(ee, ss);
	calc_mags(ee, ss);
	find_means(ee, ss);
	normalize(ee, ss, zero_point);
	calc_sigmas(ee, ss);

	if (debug_flag) {
		print_curves(ee, ss);
		print_sigmas(ee, ss);
	}
}


static void
init_matrix(ee, ss)
int ee, ss;
{

	if ((mag = matrix(1, ee, 1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix mag\n");
		exit(-1);
	}
	if ((weight = matrix(1, ee, 1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix weight\n");
		exit(-1);
	}
	if ((a = matrix(1, ee + ss, 1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix a\n");
		exit(-1);
	}
	if ((b = vector(1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector b \n");
		exit(-1);
	}
	if ((u = matrix(1, ee + ss, 1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix u\n");
		exit(-1);
	}
	if ((v = matrix(1, ee + ss, 1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix v\n");
		exit(-1);
	}
	if ((w = vector(1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector w \n");
		exit(-1);
	}
	if ((x = vector(1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector x \n");
		exit(-1);
	}
	if ((em = vector(1, ee)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector em \n");
		exit(-1);
	}
	if ((mag0 = vector(1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector mag0 \n");
		exit(-1);
	}
	if ((sigm0 = vector(1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigm0 \n");
		exit(-1);
	}
	if ((Mes = matrix(1, ee, 1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix Mes \n");
		exit(-1);
	}
	if ((sigMes = matrix(1, ee, 1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix sigMes \n");
		exit(-1);
	}
	if ((sigem = vector(1, ee)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigem \n");
		exit(-1);
	}
	if ((sigemn = vector(1, ee)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigemn \n");
		exit(-1);
	}
}



	/* calculate values of the sigmas for each exposure (as a whole),
	   for mean magnitude of each star, 
	   and for each observation of each star (individually).
	   This routine must be called AFTER do_solution(). 

	   The tricky part is to handle values of NODATA correctly so that
	   they don't throw averages off. */

static void
calc_sigmas(m, n)
int m, n;
{
	int i, j, num;
	double temp, top, bot, sig;
	struct s_star *s;
	struct s_starp *sp;

	/* set everything to 0.0, first of all */
	for (i = 1; i <= m; i++) {
		sigem[i] = 0.0;
		sigemn[i] = 0.0;
		for (j = 1; j <= n; j++)
			sigMes[i][j] = 0.0;
	}
	for (i = 1; i <= n; i++)
		sigm0[i] = 0.0;

	/* first find sigma(em(e)) */
	if (n >= 2) {
		for (i = 1; i <= m; i++) {
			num = 0;
/****
			for (sp = image_arr[i - 1].starp; sp != NULL; sp = sp->next)
				if (sp->star->include != 0)
					num++;
***/
			top = 0.0;
			bot = 0.0;
			for (j = 1; j <= n; j++) {
				if ((Mes[i][j] == NODATA) || (mag0[j] == NODATA))
					continue;
				num++;
				temp = (Mes[i][j] - mag0[j]);
				top += temp*temp*weight[i][j];
				bot += weight[i][j];
			}
			if (bot > 0.0) {
				sigem[i] = sqrt((n*top)/((n - 1)*bot));
				if (num > 0) {
					sigemn[i] = sigem[i]/(float) sqrt((double) num);
				}
			}
			else {
				sigem[i] = NODATA;	
				sigemn[i] = NODATA;	
			}
		}
	}

	/* now, find sigma(m0(s)) */
	if (m >= 2) {
		for (i = 1; i <= n; i++) {
			top = 0.0;
			bot = 0.0;
			if (star_arr[i - 1].variable == 0) {
				for (j = 1; j <= m; j++) {
					if ((Mes[j][i] == NODATA) || (mag0[i] == NODATA) ||
										(image_arr[j - 1].include == 0))
						continue;
					temp = (Mes[j][i] - mag0[i]);
					top += temp*temp*weight[j][i];
					bot += weight[j][i];
				}
			}
			else {
				for (s = star_arr[i - 1].stars; s != NULL; s = s->next) {
					j = s->image->index + 1;
					if ((Mes[j][i] == NODATA) || (mag0[i] == NODATA) ||
										(image_arr[j - 1].include == 0))
						continue;
					temp = (Mes[j][i] - mag0[i]);
					top += temp*temp*s->weight;
					bot += s->weight;
				}
			}
			if (bot > 0.0)
				sigm0[i] = sqrt((m*top)/((m - 1)*bot));
			else
				sigm0[i] = NODATA;
		}
	}

	/* now we can finally calculate sigma(M(e,s)) */
	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			temp = find_sigm(n, i, j);
			if (temp == NODATA)
				sigMes[i][j] = NODATA;
			else if (sigemn[i] != NODATA)
				sigMes[i][j] = sqrt(temp*temp + sigemn[i]*sigemn[i]);
			else
				sigMes[i][j] = temp;
		}
	}
}

print_sigmas(m, n)
int m, n;
{
	int i, j;
	double temp, top, bot, sig;

	/* first find sigma(em(e)) */
	printf("   Sigma of Exposures: image #, sigem, sigemn\n");
	for (i = 1; i <= m; i++) {
		printf(" %4d  %5.3f  %5.3f\n", image_arr[i - 1].index, sigem[i], sigemn[i]); 
	}
	printf("\n");

	/* now, find sigma(m0(s)) */
	printf("  Sigma of stellar mean mags: star #, m0, sig(m0) \n");
	for (i = 1; i <= n; i++) {
		printf("  %6d %6.3f %5.3f  \n", star_arr[i - 1].id, mag0[i], sigm0[i]); 
	}
	printf("\n");
}


	/* calculate the 'em' matrix - the overal shift of each image's 
	   magnitudes to bring them close to the mean values. */

static void
calc_em(m, n)
int m, n;
{
	int i, j;	
	
	for (i = 1; i <= m; i++) {
		if (image_arr[i - 1].include)
			em[i] = x[i];
		else
			em[i] = NODATA;
	}
}

	/* adjust the instrumental magnitudes for each star, adding the
	   results of the matrix inversion. */

static void
calc_mags(m, n)
int m, n;
{
	int i, j;	
	
	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			if ((mag[i][j] != NODATA) && (image_arr[i - 1].include != 0))
				Mes[i][j] = mag[i][j] - x[i];
			else
				Mes[i][j] = NODATA;
		}
	}
}


	/* print out light curves for each star */

print_curves(m, n)
int m, n;
{
	int i, j;	
	
	printf("\n");
	for (i = 1; i <= n; i++) {
		printf("star   %2d: ", i);
		for (j = 1; j <= m; j++) {
			printf("%6.3f %6.3f  ", Mes[j][i], sigMes[j][i]);
		}
		printf("\n");
	}
}



	/* set to zero all elements of vector p which are less than
	   'tolerance' times the maximum element of p */

static void
zero_vector(p, m, tolerance)
int m;
float *p, tolerance;
{
	int i, j;
	float max, min;

	max = 0.0;
	for (i = 1; i <= m; i++) {
		if (p[i] > max)
			max = p[i];
	}
	min = tolerance*max;
	for (i = 1; i <= m; i++) {
		if (p[i] < min)
			min = 0.0;
	}
}	



	/* print a vector (which goes 1-m) in a nice format */

static void
print_vector(p, m)
int m;
float *p;
{
	int i, j;

	for (i = 1; i <= m; i++) {
		printf("%9.3f ", p[i]);
	}
	putchar('\n');
}

	/* print a matrix (which goes 1-m, 1-n) in a nice format */

static void
print_matrix(p, m, n)
float **p;
int m, n;
{
	int i, j;

	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			printf("%9.3f ", p[i][j]);
		}
		putchar('\n');
	}
}

	/* fill up the vector B, which is computed from observed magnitudes
	   and weight */

static void 
fill_bvec(b, m, n)
float *b;
int m, n;
{
	int i, j;
	float x;

	for (i = 1; i <= m; i++) {
		x = 0;
		for (j = 1; j <= n; j++) {
			x += weight[i][j]*mag[i][j];
		}
		b[i] = x;
	}
	for (i = 1; i <= n; i++) {
		x = 0;
		for (j = 1; j <= m; j++) {
			x += weight[j][i]*mag[j][i];
		}
		b[i + m] = x;
	}
}


	/* read in the magnitude and uncertainty data */

static void
read_data(m, n)
int m, n;
{
	int i, j;
	float temp;
	struct s_star *star;

	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			mag[i][j] = 0.0;
			weight[i][j] = 0.0;
		}
	}

	for (j = 1; j <= n; j++) {
		star = star_arr[j - 1].stars;	
		while (star != NULL) {
			i = star->image->index + 1;	
			mag[i][j] = (float) star->mag;
			if ((star->include != 0) && (star->image->include != 0) &&
			  				(star_arr[j - 1].include != 0) && 
							(star_arr[j - 1].variable != 1)) {
				weight[i][j] = (float) star->weight;
			}
			star = star->next;
		}
	}

}

	/* fill up the big matrix with weight data */

static void
fill_matrix(p, m, n)
float **p;
int m, n;
{
	int i, j;
	float x;

	for (i = 1; i <= m + n; i++)
		for (j = 1; j <= m + n; j++) 	
			p[i][j] = 0.0;

	/* fill up the real matrix in three steps */
	for (i = 1; i <= m; i++) {
		x = 0;
		for (j = 1; j <= n; j++)
			x += weight[i][j];
		p[i][i] = x;
	}
	for (i = 1; i <= n; i++) {
		x = 0;
		for (j = 1; j <= m; j++)
			x += weight[j][i];
		p[i + m][i + m] = x;
	}
	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			p[i][j + m] = weight[i][j];
			p[j + m][i] = weight[i][j];
		}
	}
}

	/* copy the values of one matrix to another one */

static void 
copy_matrix(p, q, m, n)
float **p, **q;
int m, n;
{
	int i, j;

	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			q[i][j] = p[i][j];
		}
	}
}

	/* add a constant to all magnitudes (and to the x[] vector as well) 
	   so that the brightest star has the given mean magnitude value */

static void
normalize(m, n, value)
int m, n;
double value;
{
	int i, j;
	float min, diff;

	min = 1.0e6;
	for (i = 1; i <= n; i++)
		if ((mag0[i] < min) && (mag0[i] != NODATA))
			min = mag0[i];
	diff = (float) (value - min);

	for (i = 1; i <= n; i++) {
		mag0[i] += diff;
		for (j = 1; j <= m; j++)
			if (mag[j][i] != NODATA)
				Mes[j][i] += diff;
	}
			
}

	/* find the mean magnitude of each star, either
	        a. by simply looking in the 'x' vector, for 
	           stars with non-zero weights
	   or
	        b. by calculating it from the weighted average of
	           (corrected) mag on each night 

	   if there is no information for a star, use the NODATA value
	   as its mean mag */

static void
find_means(m, n)
int m, n;
{
	int i, index;
	double top, bot;
	struct s_star *star;

	for (i = 1; i <= n; i++) {
		if ((star_arr[i - 1].include != 0) && (star_arr[i - 1].variable == 0)) {
			mag0[i] = x[i + m];
		}
		else if (star_arr[i - 1].variable != 0) {
			top = 0.0;
			bot = 0.0;
			star = star_arr[i - 1].stars;
			while (star != NULL) {
				index = star->image->index + 1;
				if (Mes[index][i] == NODATA) {
					star = star->next;
					continue;
				}
				top += Mes[index][i]*star->weight;
				bot += star->weight;
				star = star->next;
			}
			if (bot > 0.0)
				mag0[i] = top/bot;
			else
				mag0[i] = NODATA;
		}
	}
}


	/* return an estimate of the error sigma(m(e,s)) for stars s in 
	   image e, which is only PART of the total error estimate
	   sigma(M(e,s)).  find this value in two ways:

	               1. try to find the average value of sigma(m0) for
	                          m0 ~= m(e,s)
	
	               2. if impossible, then use a value calculated
	                  from the weight of the value for m(e,s)
	                  since it is really based on PHOT's estimate:
	                       sig(m(e,s)) = 1.0/sqrt(weight(e,s))

	   if neither is possible, return NODATA */

static float
find_sigm(n, e, s)
int n, e, s;
{
	int i, num;
	float dist, sum, wt, ret;

	/* first, try to find a mean m0 whose magnitude is close to 
	   m(e,s).  use a weighted average of the nearby sigm0. */
	sum = 0.0;
	wt = 0.0;
	num = 0;
	for (i = 1; i <= n; i++) {
		if ((dist = fabs(mag0[i] - Mes[e][s])) < MAG_MATCH) {
			if (dist < TINY)
				dist = TINY;
			sum += sigm0[i]/dist;
			wt += 1.0/dist;
			num++;
		}
	}
	if (num > 0) {
		ret = sum/wt;
		return(ret);
	}

	/* second best: use the weight of instrumental mag(e,s) as an
	   estimator */
	if (weight[e][s] > 0.0) {
		ret = 1.0/sqrt(weight[e][s]);
		return(ret);
	}

	/* nothing left to do but return NODATA */
	ret = NODATA;
	return(ret);
}


	/* write output about images ONLY to the given file 
	   (or stdout, if the given string is ""). */

int 
write_images(file)
char *file;
{
	int i;
	FILE *fp;
	struct s_image *im;

	if (strcmp(file, "") == 0)
		fp = stdout;
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		exit(-1);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		exit(-1);
	}

	for (i = 0; i < num_images; i++) {
		if (image_arr[i].include == 0)
			continue;
		im = &(image_arr[i]);
		fprintf(fp, "%-4d %10s %10s %6.1lf %4.2lf %8.5lf %11.5lf ",
			im->index, im->name, im->filter, im->exptime, im->airmass,
			im->ut, im->mjd);
		fprintf(fp, "%5.3f %5.3f %5.3f\n", 
							em[i + 1], sigem[i + 1], sigemn[i + 1]);
	}
	if (fp != stdout)
		fclose(fp);
}

	/* write information about stars ONLY, many lines per star, to 
	   the given file (or stdout if file == "").  */

int
write_stars(file)
char *file;
{
	int i, index;
	FILE *fp;
	struct s_allstar *as;
	struct s_image *im;
	struct s_star *star;

	if (strcmp(file, "") == 0)
		fp = stdout;
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		exit(-1);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		exit(-1);
	}

	for (i = 0; i < num_ids; i++) {
		if (star_arr[i].include == 0)
			continue;
		as = &(star_arr[i]);
		star = as->stars;
		while (star != NULL) {
			im = star->image;
			if (im->include != 0) {
				index = star->image->index;
				fprintf(fp, "%-6d %6.2lf %6.2lf %4d %11.5lf ",
						as->id, as->xc, as->yc, index, im->mjd);
				fprintf(fp, " %6.3f %6.3f %5.3f\n", mag0[i + 1], 
						Mes[index + 1][i + 1], sigMes[index + 1][i + 1]);
			}
			star = star->next;
		}
	}
	if (fp != stdout)
		fclose(fp);
}

	/* write information about sigma0 vs. mag0 to the given file
	   (or stdout if the given file == "") */

int
write_sigma(file)
char *file;
{
	int i;
	FILE *fp;

	if (strcmp(file, "") == 0)
		fp = stdout;
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		exit(-1);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		exit(-1);
	}

	for (i = 0; i < num_ids; i++) {
		if (star_arr[i].include != 0) 
			fprintf(fp, "%6d %6.3f %6.3f\n", star_arr[i].id, 
						mag0[i + 1], sigm0[i + 1]);
	}

	if (fp != stdout)
		fclose(fp);
}

