#include <stdio.h>
#include <time.h>
#include <math.h>
#include "bait.h"
#include "pf.h"

	/* this file contains routines for determining various astronomical 
	   times and dates */

void set_today( /* int day, month, year */ );
void get_today( /* int *day, *month, *year */ );
void set_ut( /* double ut */ );
void get_ut( /* double ut */ );
void get_jd( /* int day, month, year, double ut, *jd */ );
int get_lst( /* double jd, *lst */ );
int lst2ut( /* double jd, lst, *ut */ );
void ut2gst( /* double jd, ut, *gst */ );
int ut2lst( /* double jd, ut, *lst */ );
void gst2ut( /* double jd, gst, *ut */ );

#ifdef VERBOSE
extern char *progname;
#endif

static int user_flag = 0;
static int user_day, user_month, user_year;
static int ut_flag = 0;
static double user_ut;

	/* set the values of the "current" day, month, and year */

void
set_today(day, month, year)
int day, month, year;
{
	user_flag = 1;
	user_day = day;
	user_month = month;
	user_year = year;
}

	/* put the values of the current Greenwich day, month and year into 
	   the passed parameters */

void
get_today(day, month, year)
int *day, *month, *year;
{
	long t, time();
	struct tm *tm;

	if (user_flag == 1) {
		*day = user_day;
		*month = user_month;
		*year = user_year;
	}
	else {
		t = time((long *) 0);
		tm = gmtime(&t);
		*day = tm->tm_mday;
		*month = tm->tm_mon + 1;		/* gmtime puts month in range 0-11 */
		*year = tm->tm_year + 1900;
	}
}

	/* set the ut - used only when simulating observations */

void
set_ut(ut)
double ut;
{
	user_ut = ut;
	ut_flag = 1;
}

	/* place the current UT, in fractional hours (i.e. 0.00 -- 23.99), 
	   into the passed variable. */

void
get_ut(ut)
double *ut;
{
	long t, time();
	struct tm *tm;

	if (ut_flag == 1)
		*ut = user_ut;
	else {
		t = time((long *) 0);
		tm = gmtime(&t);
		*ut = (double)tm->tm_hour + tm->tm_min/60.0 + tm->tm_sec/3600.0;
	}
}
	
	/* place the fractional Julian Date, which corresponds to the given
	   day, month, year and UT time, into the passed variable */

#define BASE     2433282.5         	/* JD of January 1.0 1950 */

void
get_jd(day, month, year, ut, jd)
int day, month, year;
double ut, *jd;
{
	double d;
	
	d = (double) since_1950(day, month, year);
	*jd = d + BASE + (ut/24.0);
}

	/* place the LST for the given Julian Date into the passed variable. 
	   uses the procedure 	
	   given by Duffett-Smith's "Practical Astronomy with your Pocket
	   Calculator", p. 17 and 20. The observatory longitude is read
	   out of the configuration file - note that it is ALREADY in the
	   form of "hours west of Greenwich".

	   the LST is set in the range (0.0-23.99). if all went well,
	   0 is returned; a return value of -1 indicates some error. */

int
get_lst(jd, lst)
double jd, *lst;
{
	double j, s, t, t0, u, longitude;

	j = floor(jd + 0.5) - 0.5;
	s = j - 2451545.0;
	t = s/36525.0;
	t0 = 6.697374558 + 2400.051336*t + 0.000025862*t*t;
	while (t0 > 24.0)
		t0 -= 24.0;
	while (t0 < 0.0)
		t0 += 24.0;
	u = (jd - j)*24.0;
	u *= 1.002737909;
	t0 += u;
	while (t0 > 24.0)
		t0 -= 24.0;
	while (t0 < 0.0)
		t0 += 24.0;

	if (read_longitude(&longitude) < 0) {
#ifdef VERBOSE
		fprintf(stderr, "%s.lst: can't read longitude\n", progname);
#endif
		return(-1);
	}

	t0 -= longitude;
	while (t0 > 24.0)
		t0 -= 24.0;
	while (t0 < 0.0)
		t0 += 24.0;

	*lst = t0;
	return(0);
}
	
	/* convert the given Local Sidereal Time on the given Julian Date to 
	   Universal Time. This routine always picks the smaller of the two 
	   possible UT values for LST values between (0h, 0m) and (0h, 4m). 

	   if for some reason the routine can't read the observatory longitude
	   out of the configuration file, it returns -1.  Otherwise, the UT
	   in the range (0.0 - 23.9333) is placed in the passed parameter, 
	   and 0 is returned. */

int
lst2ut(jd, lst, ut)
double jd, lst, *ut;
{
	double j, s, t, t0, gst, longitude;
	
	/* first, convert LST to Greenwich mean ST */
	if (read_longitude(&longitude) < 0) {
#ifdef VERBOSE
		fprintf(stderr, "%s.lst2ut: can't read longitude\n", progname);
#endif
		return(-1);
	}
	gst = lst + longitude;
	while (gst > 24.0)	
		gst -= 24.0;
	while (gst < 0.0)
		gst += 24.0;

	/* now convert from GMST to UT */
	j = floor(jd + 0.5) - 0.5;
	s = j - 2451545.0;
	t = s/36525.0;
	t0 = 6.697374558 + 2400.051336*t + 0.000025862*t*t;
	while (t0 > 24.0)
		t0 -= 24.0;
	while (t0 < 0.0)
		t0 += 24.0;
	
	gst -= t0;
	while (gst > 24.0)	
		gst -= 24.0;
	while (gst < 0.0)
		gst += 24.0;
	*ut = gst*0.9972695663;

	return(0);
}

	/* convert from UT (fractional hours) on the given Julian Date to GST 
	   (again, fractional hours). */

void
ut2gst(jd, ut, gst)
double jd, ut, *gst;
{
	double j, s, t, t0;

	j = floor(jd + 0.5) - 0.5;
	s = j - 2451545.0;
	t = s/36525.0;
	t0 = 6.697374558 + 2400.051336*t + 0.000025862*t*t;
	while (t0 > 24.0)
		t0 -= 24.0;
	while (t0 < 0.0)
		t0 += 24.0;
	t0 += ut*1.002737909;
	while (t0 > 24.0)
		t0 -= 24.0;
	while (t0 < 0.0)
		t0 += 24.0;
	*gst = t0;
}

	/* convert from UT (fractional hours) to Local Sidereal Time. return 0
	   if all went well, or -1 if there was a problem reading the longitude
	   from the CONFIG file. */

int
ut2lst(jd, ut, lst)
double jd, ut, *lst;
{
	double longitude, gst;

	if (read_longitude(&longitude) < 0) {
#ifdef VERBOSE
		fprintf(stderr, "%s.ut2lst: can't read longitude\n", progname);
#endif
		return(-1);
	}
	ut2gst(jd, ut, &gst);
	*lst = gst - longitude;		/* longitude is in fractional hours already */
	if (*lst < 0.0)
		*lst += 24.0;

	return(0);
}



	/* convert from GST (fractional hours) on the given JD to UT
	   (also fractional hours). the output UT is in the range 
	   (0.0 - 23.9333); in case of ambiguity, the smaller UT value is
	   always chosen. */

void
gst2ut(jd, gst, ut)
double jd, gst, *ut;
{
	double j, s, t, t0;
	
	/* now convert from GST to UT */
	j = floor(jd + 0.5) - 0.5;
	s = j - 2451545.0;
	t = s/36525.0;
	t0 = 6.697374558 + 2400.051336*t + 0.000025862*t*t;
	while (t0 > 24.0)
		t0 -= 24.0;
	while (t0 < 0.0)
		t0 += 24.0;
	
	gst -= t0;
	while (gst > 24.0)	
		gst -= 24.0;
	while (gst < 0.0)
		gst += 24.0;
	*ut = gst*0.9972695663;
}

