
	/* a program that allows one to edit FITS headers, by the following 
	   means:
	            1. given a FITS file name, the program figures out how 
	               long the header is, and

	            2. copies the entire header (up to 2880-byte boundary)
	               into a temporary file, then

	            3. changes all the trailing blanks in each line into a 
	               single newline ('\n') character, then

	            4. forks, and exec's 'vi' on the temporary file.

	            5. when 'vi' editing finished, converts the temporary file
	               lines back into 80-char lines, with blanks and no 
	               newlines.

	            6. writes the header back into the original FITS file.

	   thus, the user cannot add more header lines than will fit into the
	   current number of 2880-byte header blocks, but otherwise can do
	   a fair bit of editing. */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#define FITSLINE       80	/* length of FITS header lines */
#define NUM_LINES      36	/* number of lines in a header block */
#define BUFLEN        255

main(argc, argv)
int argc;
char *argv[];
{
	int i, header_line, fd, done;
	char line[BUFLEN+1], tempname[30], filename[30];
	FILE *fp;

	if (argc != 2) {
		fprintf(stderr, "usage: vifits filename\n");
		exit(-1);
	}
	strcpy(filename, argv[1]);

	if ((fd = open(filename, O_RDWR)) < 0) {
		fprintf(stderr, "can't open file %s \n", filename);
		exit(-1);
	}
	strcpy(tempname, "/tmp/vifXXXXXX");
	if (mktemp(tempname) == NULL) {
		fprintf(stderr, "can't create temporary filename \n");
		exit(-1);
	}
	if ((fp = fopen(tempname, "w")) < 0) {
		fprintf(stderr, "can't open temporary file %s\n", tempname);
		exit(-1);
	}

	/* copy the FITS header into a file for editing */
	done = 0;
	header_line = 0;
	while (!done) {
		header_line++;
		if (read(fd, line, FITSLINE) != FITSLINE) {
			fprintf(stderr, "error reading file near header line %d\n",
					header_line);
			exit(-1);
		}
		if (strncmp(line, "END", 3) == 0)
			done = 1;
		strip_blanks(line);
		fprintf(fp, "%s", line);
	}
	while (header_line % NUM_LINES != 0) {
		fprintf(fp, "\n");
		header_line++;
	}
	fclose(fp);

	/* now spawn 'vi' to edit the file */
	if (fork() == 0) { 
		execlp("vi", "vi", tempname, (char *)0);
	}
	else
		wait((int *) 0);

	/* and now, copy the edited header back into its FITS file, in FITS
	   format */
	if ((fp = fopen(tempname, "r")) == NULL) {
		fprintf(stderr, "can't re-open temporary file %s\n", tempname);
		exit(-1);
	}
	for (i = 0; i < header_line; i++) {
		fgets(line, BUFLEN, fp);
		add_blanks(line);
		lseek(fd, i*FITSLINE, 0);
		if (write(fd, line, FITSLINE) != FITSLINE) {
			fprintf(stderr, "error writing line %d back to FITS header \n",
					i);
			exit(-1);
		}
	}
	close(fd);
	fclose(fp);
	unlink(tempname);
}

	/* replace all the trailing blanks in the given line (from character 
	   position 80 backwards) with a single newline character, then put 
	   a NULL character after the newline character. */

strip_blanks(line)
char *line;
{
	int i;

	for (i = FITSLINE - 1; i > 0; i--) {
		if (line[i] != ' ')
			break;
	}
	line[i + 1] = '\n';
	line[i + 2] = '\0';
}

	/* pad a string out to 80 characters long with blanks.
	   If the line is longer than 80 characters, do nothing. */

add_blanks(line)
char *line;
{
	int i, j;

	if ((j = strlen(line)) >= FITSLINE)
		return(0);
	for (i = j - 1; i < FITSLINE; i++) {
		line[i] = ' ';
	}
	line[i] = ' ';
	line[i + 1] = '\0';
}
