
	/* routines herein check and modify the observing windows for one or
	   more objects; each object may have one or two windows in JD, and
	   these routines try to find the intersections of windows. Also
	   herein are very similar routines that adjust observing windows
	   so that the moon does NOT intersect them (if it's too bright). */

#include <stdio.h>
#include "bait.h"
#include "list.h"

static int outside( /* struct s_object *obj, int i, 
                       struct s_group *grp, int j */ );
static int time_fits( /* struct s_object *obj, 
                         struct s_group *grp, int j */ );
static void shrink_window( /* struct s_object *obj, int i, 
                              struct s_group *grp, int j */ );
int fit_object( /* struct s_object *obj, struct s_group *grp */ );
int avoid_moon( /* double moon_rise, moon_set, struct s_group *grp */ );
static void moon_shrink( /* double moon_rise, moon_set, 
                            struct s_group *grp, int i */ );

#ifdef VERBOSE
extern char *progname;
#endif

	/* return 1 if the first window start,end pair is outside
	   the seconds start,end pair, or 0 if there is some overlap */

static int
outside(obj, i, grp, j)
int i, j;
struct s_object *obj;
struct s_group *grp;
{
	return((obj->jd_start[i] > grp->jd_end[j]) || 
	       (obj->jd_end[i] < grp->jd_start[j]));
}

	/* call this routine to see if the object's exposure time will fit
	   inside the window (when added to the sum of all other objects
	   already in the group). return 1 if the object fits, or 0 if not. */

static int
time_fits(obj, grp, j)
int j;
struct s_object *obj;
struct s_group *grp;
{
	double win_len, exp_len;

	win_len = grp->jd_end[j] - grp->jd_start[j];
	exp_len = (grp->exp_time + obj->exp_time + SLOP_TIME)/86400.0;
	if (win_len <= exp_len)
		return(0);
	else
		return(1);
}

	/* given two windows, shrink the second, if
	   necessary, so that the first is larger than or equal to it. 
	   Also set the 'win_used[j]' flag of the group to 1 if the 
	   object fits within the group's window j, or 0 if not. */

static void
shrink_window(obj, i, grp, j)
int i, j;
struct s_object *obj;
struct s_group *grp;
{
	if ((outside(obj, i, grp, j)) || (!time_fits(obj, grp, j)))
		grp->win_used[j] = 0;
	else {	
		grp->win_used[j] = 1;
		if (obj->jd_start[i] > grp->jd_start[j])
			grp->jd_start[j] = obj->jd_start[i];
		if (obj->jd_end[i] < grp->jd_end[j])
			grp->jd_end[j] = obj->jd_end[i];
	}
}

	/* try to fit the given object into the given group, modifying the
	   number of and ends of the group's windows so that the new object is
	   included. If only one window is still viable, make sure that it is
	   put in the first position. 

	   return 0 if the new object could be fitted into the group, or -1 if
	   it could not. */

	/* this macro moves the second window into the first window's slot */
#define SLIDE(grp)  grp->jd_start[0] = grp->jd_start[1]; \
                    grp->jd_end[0] = grp->jd_end[1]; \
                    grp->win_used[0] = 1; \
                    grp->win_used[1] = 0;

int
fit_object(obj, grp)
struct s_object *obj;
struct s_group *grp;
{
	int code;

	/* case A: both object and group have only one window */
	if ((obj->vis_flag == 1) && (grp->vis_flag == 1)) {
		shrink_window(obj, 0, grp, 0);
		if (grp->win_used[0] == 0) {
#ifdef VERBOSE
			fprintf(stderr, "%s.fit_object: can't fit obj into grp, case A\n", 
					progname);
#endif
			grp->vis_flag = 0;
			return(-1);
		}
		return(0);
	}

	/* case B: object has only one window, but group has two windows */
	if ((obj->vis_flag == 1) && (grp->vis_flag == 2)) {
		shrink_window(obj, 0, grp, 0);
		shrink_window(obj, 0, grp, 1);
		if ((grp->vis_flag = grp->win_used[0] + grp->win_used[1]) == 0) {
#ifdef VERBOSE
			fprintf(stderr, "%s.fit_object: can't fit obj into grp, case B\n",
					progname);
#endif
			return(-1);
		}	
		else if ((grp->vis_flag == 1) && (grp->win_used[0] == 0)) {
			SLIDE(grp);
		}
	}

	/* case C: object has two windows, but group has only one window */
	if ((obj->vis_flag == 2) && (grp->vis_flag == 1)) {

		/* first, duplicate the group's single window in the other slot */
		grp->jd_start[1] = grp->jd_start[0];
		grp->jd_end[1] = grp->jd_end[0];
		grp->win_used[1] = 1;
		shrink_window(obj, 0, grp, 0);
		shrink_window(obj, 0, grp, 1);
		if ((grp->vis_flag = grp->win_used[0] + grp->win_used[1]) == 0) {	
#ifdef VERBOSE
			fprintf(stderr, "%s.fit_object: can't fit object into grp, case C\n",
					progname);
#endif
			return(-1);
		}

		/* if only the second window is still open, move it over to the first */
		if ((grp->vis_flag == 1) && (grp->win_used[0] = 0)) {
			SLIDE(grp);
		}
		return(0);
	}

	/* case D: both the object and the group have two windows. In this case it
	   IS possible for three windows of overlap to exist, but in such a case
	   I will ignore the third: at most two windows of overlap come out of
	   this section of case. */
	code = 8*(!outside(obj, 0, grp, 0) && time_fits(obj, grp, 0)) + 
	       4*(!outside(obj, 0, grp, 1) && time_fits(obj, grp, 1)) +
	       2*(!outside(obj, 1, grp, 0) && time_fits(obj, grp, 0)) +   
	         (!outside(obj, 1, grp, 1) && time_fits(obj, grp, 1));
	switch (code) {
	case 0:
		grp->vis_flag = 0;
		grp->win_used[0] = 0;
		grp->win_used[1] = 0;
#ifdef VERBOSE
		fprintf(stderr, "%s.fit_object: can't fit object into grp, case D\n",
				progname);
#endif
		return(-1);
		break;
	case 1:
		shrink_window(obj, 1, grp, 1);
		grp->vis_flag = 1;
		SLIDE(grp);
		break;
	case 2:
		shrink_window(obj, 1, grp, 0);
		grp->vis_flag = 1;	
		grp->win_used[1] = 0;
		break;
	case 3:
		shrink_window(obj, 1, grp, 0);
		shrink_window(obj, 1, grp, 1);
		grp->vis_flag = 2;
		break;
	case 4:
		shrink_window(obj, 0, grp, 1);
		grp->vis_flag = 1;
		SLIDE(grp);
		break;
	case 5:
		grp->jd_start[0] = grp->jd_start[1];
		grp->jd_end[0] = grp->jd_end[1];
		shrink_window(obj, 0, grp, 0);
		shrink_window(obj, 1, grp, 1);
		grp->vis_flag = 2;
		break;
	case 8:	
		shrink_window(obj, 0, grp, 0);
		grp->vis_flag = 1;
		grp->win_used[1] = 0;
		break;
	case 9:
		shrink_window(obj, 0, grp, 0);
		shrink_window(obj, 1, grp, 1);
		grp->vis_flag = 2;
		break;
	case 10:
		grp->jd_start[1] = grp->jd_start[0];
		grp->jd_end[1] = grp->jd_end[0];
		shrink_window(obj, 0, grp, 0);
		shrink_window(obj, 1, grp, 1);
		grp->vis_flag = 2;
		break;
	/* these are the cases where I just plain ignore one region of overlap */
	case 11:	/* in this case, [obj, 1, grp, 0] intersection ignored */ 
	case 13:	/* in this case, [obj, 0, grp, 1] intersection ignored */ 
		shrink_window(obj, 0, grp, 0);
		shrink_window(obj, 1, grp, 1);
		grp->vis_flag = 2;
		break;
	case 12:
		shrink_window(obj, 0, grp, 0);
		shrink_window(obj, 0, grp, 1);
		grp->vis_flag = 2;
		break;	
	case 6:	
	case 7:
	case 14:
	case 15:
#ifdef VERBOSE
		fprintf(stderr, "%s.fit_window: impossible case, code %d, occurs!\n",
				progname, code);
#endif
		return(-1);
	}
	return(0);
}
		
	/* similar routines for folding in the moon rise/set times into a
	   group's observing window */

	/* try to keep the moon out of the given group, modifying the
	   number of and ends of the group's windows so that the moon ISN'T
	   included. If only one window is still viable, make sure that it is
	   put in the first position. 

	   As before, it is POSSIBLE for a situation to arise in which an
	   object does have three windows, but this code will choose only
	   two of the three and discard one. Tough luck - although this 
	   moon routine does try to choose the longer windows...

	   return 0 if the group can still be observed at all, or -1 if
	   it can not. */

int
avoid_moon(moon_rise, moon_set, grp)
double moon_rise, moon_set;
struct s_group *grp;
{
	int code;

	/* wierd case: the group has no observing window already. This should
	   never happen, but return -1 just to be consistent. */
	if (grp->vis_flag == 0) {
#ifdef VERBOSE
		fprintf(stderr, "%s.avoid_moon: impossible case!? vis_flag = 0 \n",
				progname);
#endif
		return(-1);
	}

	/* case M1: the group has only one window */
	if (grp->vis_flag == 1) {
		moon_shrink(moon_rise, moon_set, grp, 0);
		if (grp->win_used[0] == 0) {
#ifdef VERBOSE
			fprintf(stderr, "%s.avoid_moon: can't avoid moon, case M1\n", 
					progname);
#endif
			grp->vis_flag = 0;
			return(-1);
		}
		return(0);
	}

	/* case M2: group has two windows */
	if (grp->vis_flag == 2) {
		moon_shrink(moon_rise, moon_set, grp, 0);
		moon_shrink(moon_rise, moon_set, grp, 1);
		if ((grp->vis_flag = grp->win_used[0] + grp->win_used[1]) == 0) {
#ifdef VERBOSE
			fprintf(stderr, "%s.avoid_moon: can't avoid moon, case M2\n",
					progname);
#endif
			return(-1);
		}	
		else if ((grp->vis_flag == 1) && (grp->win_used[0] == 0)) {
			SLIDE(grp);
		}
		return(0);
	}

	/* should never get here */
#ifdef VERBOSE
	fprintf(stderr, "%s.avoid_moon: impossible case ?! vis_flag = %d > 2 \n", 
			progname, grp->vis_flag);
#endif
}

	/* given a group window and moon rise/set times, shrink the window, if
	   necessary, so that it does not intersect the moon times at all.
	   Also set the 'win_used[j]' flag of the group to 1 if the 
	   object may still be observed within some part of the window, or 0
	   if the window is completely covered by the moon. */

static void
moon_shrink(moon_rise, moon_set, grp, i)
double moon_rise, moon_set;
struct s_group *grp;
int i;
{
	if (((moon_rise < grp->jd_start[i]) && (moon_set < grp->jd_start[i])) ||
	    ((moon_rise > grp->jd_end[i]) && (moon_set > grp->jd_end[i])) ||
	    ((moon_rise > grp->jd_end[i]) && (moon_set < grp->jd_start[i]))) {
		grp->win_used[i] = 1;
	}
	else if (((moon_rise < grp->jd_start[i]) && (moon_set > grp->jd_end[i])) ||
	    ((moon_rise < grp->jd_start[i]) && (moon_set < moon_rise))) {
		grp->win_used[i] = 0;
	}
	else if ((moon_rise < grp->jd_start[i]) && (moon_set < grp->jd_end[i])) {
		grp->win_used[i] = 1;
		grp->jd_start[i] = moon_set;
	}
	else if ((moon_rise < grp->jd_end[i]) && (moon_set < grp->jd_start[i])) {
		grp->win_used[i] = 1;
		grp->jd_end[i] = moon_rise;
	}
	else if ((moon_rise < grp->jd_end[i]) && (moon_set < grp->jd_end[i])) {
		if (moon_set < moon_rise) {
			/* this is easy - just squeeze in both ends of the window */
			grp->win_used[i] = 1;
			grp->jd_start[i] = moon_set;
			grp->jd_end[i] = moon_rise;
		}
		else {
		/* this is a tricky case - we have two possible windows from
		   which to choose: before moonrise or after moonset. Let's do
		   the following: if there is another window being used, just
		   choose the one of these two which is longer in time. Otherwise,
		   set the group's first window to before moonrise, and second to
		   after moonset */
			grp->win_used[i] = 1;
			if (grp->vis_flag == 2) {
				if ((moon_rise - grp->jd_start[i]) > (grp->jd_end[i] - moon_set))
					grp->jd_end[i] = moon_rise;
				else
					grp->jd_start[i] = moon_set;
			}
			else {
				if (i == 0) {
					grp->win_used[1] = 1;
					grp->jd_start[1] = moon_set;
					grp->jd_end[1] = grp->jd_end[0];
					grp->jd_end[0] = moon_rise;
				}
				else {
					grp->win_used[0] = 1;
					grp->jd_start[0] = grp->jd_start[1];
					grp->jd_end[0] = moon_rise;
					grp->jd_start[1] = moon_set;
				}
			}
		}
	}
	else if ((moon_rise < grp->jd_end[i]) && (moon_set > grp->jd_end[i])) {
		grp->win_used[i] = 1;
		grp->jd_end[i] = moon_rise;
	}
	else {
		/* we should never get here, having taken care of all possible cases 
		   above.  Just print out a warning message, but take no action. */
#ifdef VERBOSE
		fprintf(stderr, "%s.moon_shrink: impossible case ?! \n", progname);
#endif
	}
}
